/* Sudoku Game
   original by Rathi

   edited by Alex Gustow
   (try to use "auto-zoom technique" - maybe I can?.. yes I can!)
*/

#include "minigui.ch"

****************************
Function Main

  local bColor
  local nRandom := 0
  local nGuess := 0
  local cTime := time()
  local cTitle := "HMG Sudoku"

  // GAL - coefs for "auto-zooming"
  private gkoefh := GetDesktopWidth()/1024, ;
          gkoefv := GetDesktopHeight()/768
  private gw := 50*gkoefh                  // for grid columns widths
  //

  private aPuzzles := ImportFromTxt("sudoku.csv")
  private aSudoku := { { 0, 0, 0, 2, 0, 3, 8, 0, 1 }, ;
                       { 0, 0, 0, 7, 0, 6, 0, 5, 2 }, ;
                       { 2, 0, 0, 0, 0, 0, 0, 7, 9 }, ;
                       { 0, 2, 0, 1, 5, 7, 9, 3, 4 }, ;
                       { 0, 0, 3, 0, 0, 0, 1, 0, 0 }, ;
                       { 9, 1, 7, 3, 8, 4, 0, 2, 0 }, ;
                       { 1, 8, 0, 0, 0, 0, 0, 0, 6 }, ;
                       { 7, 3, 0, 6, 0, 1, 0, 0, 0 }, ;
                       { 6, 0, 5, 8, 0, 9, 0, 0, 0 }  }
  private aOriginal := {}

  // GAL
  if file("Help.chm")
    set helpfile to 'Help.chm'
  endif
  //

  if len(aPuzzles) > 0
    nRandom := val("0."+substr(cTime,8,1)+substr(cTime,5,1)+substr(cTime,8,2))
    nGuess := int(nRandom * len(aPuzzles))  
    if nGuess == 0
      nGuess := 1
    endif
    aSudoku := aclone(aPuzzles[nGuess])
    cTitle := "HMG Sudoku"+" Game no: "+alltrim(str(nGuess))+" of "+alltrim(str(len(aPuzzles)))
  endif   

  bColor := { || SudokuBackColor() }

  aOriginal := aclone(aSudoku)

  //define window Sudoku at 0,0 width 500 height 550 main title cTitle

  // GAL
  define window Sudoku ;
    at 0,0 ;
    width 500*gkoefh height 560*gkoefv ;  /* 550*gkoefv */
    main ;
    title cTitle ;
    nomaximize ;    /* GAL */
    nosize          /* GAL */

    define grid Square
      row 10
      col 10
      width 460*gkoefh
      height 445*gkoefv
      showheaders .f.
      headers {"", "", "", "", "", "", "", "", ""}
      widths  {gw, gw, gw, gw, gw, gw, gw, gw, gw}   /* all = 50 */
      justify { 2, 2, 2, 2, 2, 2, 2, 2, 2}
      allowedit .T.
      columncontrols { {"TEXTBOX", "CHARACTER", "9"}, ;
                       {"TEXTBOX", "CHARACTER", "9"}, ;
                       {"TEXTBOX", "CHARACTER", "9"}, ;
                       {"TEXTBOX", "CHARACTER", "9"}, ;
                       {"TEXTBOX", "CHARACTER", "9"}, ;
                       {"TEXTBOX", "CHARACTER", "9"}, ;
                       {"TEXTBOX", "CHARACTER", "9"}, ;
                       {"TEXTBOX", "CHARACTER", "9"}, ;
                       {"TEXTBOX", "CHARACTER", "9"} }
      fontsize 30*gkoefh
      dynamicbackcolor { bColor, bColor, bColor, ;
                         bColor, bColor, bColor, ;
                         bColor, bColor, bColor }
      columnwhen { { || entergrid() }, { || entergrid() }, ;
                   { || entergrid() }, { || entergrid() }, ;
                   { || entergrid() }, { || entergrid() }, ;
                   { || entergrid() }, { || entergrid() }, ;
                   { || entergrid() } }
      columnvalid { { || checkgrid() }, { || checkgrid() }, ;
                    { || checkgrid() }, { || checkgrid() }, ;
                    { || checkgrid() }, { || checkgrid() }, ;
                    { || checkgrid() }, { || checkgrid() }, ;
                    { || checkgrid() } }
      onchange CheckPossibleValues()     
      fontname "Arial"
      cellnavigation .T.
    end grid

    define label valid
      row 460*gkoefv
      col 10
      fontname "Arial"
      width 370*gkoefh      /* 430*gkoefh */
      fontsize 18*gkoefv
    end label

    // GAL
    if file("help.chm")

      define button Help
        row 460*gkoefv
        col 385*gkoefh
        width 50*gkoefh
        caption "Help"
        action DISPLAY HELP MAIN
      end button

      define statusbar
        statusItem 'F1 - Help' action DISPLAY HELP MAIN
      end statusbar

    endif

    define button Next
      row 460*gkoefv
      col 435*gkoefh        /* 440*gkoefh */
      width 50*gkoefh       /* 40*gkoefh */
      caption "Next"
      action NextGame()
    end button

  end window

  on key ESCAPE of Sudoku action Sudoku.release()

  // GAL
  if file("help.chm")
    on key F1 of Sudoku action DISPLAY HELP MAIN
  endif

  RefreshSudokuGrid()

  Sudoku.center
  Sudoku.activate

Return Nil

****************************
function SudokuBackColor()

  do case

    case this.cellrowindex <= 3 // first row
      do case
        case this.cellcolindex <= 3 // first col
          return {200,100,100}
        case this.cellcolindex > 3 .and. this.cellcolindex <= 6// second col
          return {100,200,100}
        case this.cellcolindex > 6 // third col
          return {100,100,200}
      endcase

    case this.cellrowindex > 3 .and. this.cellrowindex <= 6  // second row       
      do case
        case this.cellcolindex <= 3 // first col
          return {100,200,100}
        case this.cellcolindex > 3 .and. this.cellcolindex <= 6// second col
          return {200,200,100}
        case this.cellcolindex > 6 // third col
          return {100,200,100}
      endcase

    case this.cellrowindex > 6 // third row       
      do case
        case this.cellcolindex <= 3 // first col
          return {100,100,200}
        case this.cellcolindex > 3 .and. this.cellcolindex <= 6// second col
          return {100,200,100}
        case this.cellcolindex > 6 // third col
          return {200,100,100}
      endcase

  endcase

Return Nil           

****************************
function RefreshSudokuGrid()

  local i := 0
  local aLine := {}
  local aValue := sudoku.square.value

  sudoku.square.DeleteAllItems()

  if len(aSudoku) == 9
    for i := 1 to len(aSudoku)
      asize(aLine,0)
      for j := 1 to len(aSudoku[i])
        if aSudoku[i,j] > 0
          aadd(aLine, str(aSudoku[i,j],1,0))
        else
          aadd(aLine,'')
        endif
      next j
      sudoku.square.AddItem(aLine)   
    next i
  endif

  sudoku.square.value := aValue

Return Nil

****************************
function EnterGrid()

  local aValue := sudoku.square.value

  if len(aValue) > 0
    if aOriginal[ aValue[1], aValue[2] ] > 0
      return .F.
    else
      return .T.
    endif
  endif

Return .F.

****************************
function CheckGrid()

  local nRow := this.cellrowindex
  local nCol := this.cellcolindex
  local nValue := val(alltrim(this.cellvalue))
  local aLine := {}
  local i := 0
  local j := 0
  local nRowStart := (int((nRow-1)/3) * 3) + 1
  local nRowEnd := nRowstart + 2
  local nColStart := (int((nCol-1)/3) * 3) + 1
  local nColEnd := nColstart + 2

  if nValue == 0
    this.cellvalue := ''
    sudoku.valid.value := ''
    aSudoku[nRow,nCol] := 0
    return .T.
  endif

  if nValue == aSudoku[nRow,nCol]
    return .T.
  endif

  for i := 1 to 9
    if aSudoku[nRow,i] == nValue //row checking
      return .F.
    endif
    if aSudoku[i,nCol] == nValue //col checking
      return .F.
    endif
  next i

  for i := nRowStart to nRowEnd
    for j := nColStart to nColEnd
      if aSudoku[i,j] == nValue
        return .F.
      endif
    next j
  next i

  sudoku.valid.value := ''

  aSudoku[nRow,nCol] := nValue

  CheckCompletion()

Return .T.
   
****************************
function CheckCompletion()

  for i := 1 to len(aSudoku)
    for j := 1 to len(aSudoku[i])
      if aSudoku[i,j] == 0
        return Nil
      endif
    next j
  next i

  MsgInfo("Congrats! You won!")

Return Nil

****************************
function CheckPossibleValues()

  local aValue := sudoku.square.value
  local aAllowed := {}
  local cAllowed := ""
  local nRowStart := (int((aValue[1]-1)/3) * 3) + 1
  local nRowEnd := nRowstart + 2
  local nColStart := (int((aValue[2]-1)/3) * 3) + 1
  local nColEnd := nColstart + 2
  local lAllowed := .t.
  local i := 0
  local j := 0
  local k := 0

  if aValue[1] > 0 .and. aValue[2] > 0

    if aOriginal[aValue[1],aValue[2]] > 0

      sudoku.valid.value := ""
      return Nil

    else

      for i := 1 to 9

        lAllowed := .t.
        for j := 1 to 9
          if aSudoku[aValue[1],j] == i
            lAllowed := .f.
          endif
          if aSudoku[j,aValue[2]] == i
            lAllowed := .f.
          endif
        next j

        for j := nRowStart to nRowEnd
          for k := nColStart to nColEnd
            if aSudoku[j,k] == i
              lAllowed := .f.
            endif
          next k
        next j

        if lAllowed
          aadd(aAllowed,i)
        endif

      next i

      if len(aAllowed) > 0
        for i := 1 to len(aAllowed)
          if i == 1
            cAllowed := cAllowed + alltrim(str(aAllowed[i]))
          else
            cAllowed := cAllowed + ", "+ alltrim(str(aAllowed[i]))
          endif
        next i
        sudoku.valid.value := "Possible Numbers: "+cAllowed
      else
        sudoku.valid.value := "Possible Numbers: Nil"
      endif               
      return Nil

    endif

  endif

Return Nil

****************************
function ImportFromTxt( cFilename )

  local aLines := {}
  local handle := fopen(cFilename,0)
  local size1 := 0
  local size2 := 0
  local sample := 0
  local lineno := 0
  local eof1 := .F.
  local linestr := ""
  local c := ""
  local len1 := 0
  local x := 0
  local finished := .F.
  local m := 0
  local length1 := 0
  local linecount := 0
  local length2 := 0
  local aPuzzles := {}
  local aPuzzle := {}
  local aRow := {}

  if handle == -1
    return aPuzzles
  endif

  size1 := fseek(handle,0,2)
  size2 := size1

  if size1 > 65000
    sample := 65000
  else
    sample := size1
  endif

  fseek(handle,0)
  lineno := 1
  aadd(aLines,"")
  c := space(sample)
  eof1 := .F.
  linestr := ""
  len1 := 0

  do while .not. eof1
    x := 0
    x := fread(handle,@c,sample)
    len1 := len1 + sample

   if x < 1

     eof1 := .T.
     aLines[lineno] := linestr

   else

     finished := .F.

     do while .not. finished

       m := at(chr(13),c)

       if m > 0

          if m == 1

            linestr := ""
            lineno := lineno + 1
            aadd(aLines,"")
            if asc(substr(c,m+1,1)) == 10
              c := substr(c,m+2,len(c))
            else
              c := substr(c,m+1,len(c))
            endif

          else

            if len(alltrim(linestr)) > 0
              linestr := linestr + substr(c,1,m-1)
            else
              linestr := substr(c,1,m-1)
            endif

            if asc(substr(c,m+1,1)) == 10
               c := substr(c,m+2,len(c))
            else
               c := substr(c,m+1,len(c))
            endif   

            aLines[lineno] := linestr
            linestr := ""
            lineno := lineno + 1
            aadd(aLines,"")

          endif

        else

          linestr := c
          finished := .t.

        endif

      enddo

      c := space(sample)

    endif

  enddo

  fclose(handle)

  for i := 1 to len(aLines)
    x := 1
    asize(aPuzzle,0)

    for j := 1 to 9 
       asize(aRow,0)
       for k := 1 to 9
         aadd(aRow,val(alltrim(substr(aLines[i],x,1))))
         x := x + 2
       next k
       if len(aRow) == 9
         aadd(aPuzzle,aclone(aRow))
       endif   
    next j

    if len(aPuzzle) == 9
      aadd(aPuzzles,aclone(aPuzzle))
    endif   

  next i

Return aPuzzles                   

****************************
function NextGame()

  local cTime := time()
  local nRandom := 0
  local nGuess := 0

  if len(aPuzzles) > 0
    nRandom := val("0."+substr(cTime,8,1)+substr(cTime,5,1)+substr(cTime,8,2))
    nGuess := int(nRandom * len(aPuzzles))  
    if nGuess == 0
      nGuess := 1
    endif
    aSudoku := aclone(aPuzzles[nGuess])
    aOriginal := aclone(aSudoku)
    sudoku.title := "HMG Sudoku"+" Game no: "+alltrim(str(nGuess))+" of "+alltrim(str(len(aPuzzles))) 
  endif

  RefreshSudokuGrid()

Return Nil
